from sqlalchemy import and_
from sqlalchemy.orm import joinedload

from database import db, ITEMS_PER_PAGE, cache
from flask_covid19.blueprints.app_all.all_model import BlueprintDateReported, BlueprintLocationGroup
from flask_covid19.blueprints.app_all.all_model import BlueprintLocation, BlueprintFactTable, BlueprintEntity


class RkiMeldedatum(BlueprintDateReported):
    __tablename__ = 'rki_date_reported'
    __mapper_args__ = {'concrete': True}
    __table_args__ = (
        db.UniqueConstraint(
            'date_reported_import_str',
            'datum',
            'year_day_of_year',
            name="uix_rki_cases_date_reported"
        ),
    )

    def __repr__(self):
        return "%s(%s, %s, %s)" % (self.__class__.__name__,
                                   self.date_reported_import_str, self.datum.isoformat(), self.year_day_of_year)

    id = db.Column(db.Integer, primary_key=True)
    #
    date_reported_import_str = db.Column(db.String(255), nullable=False, unique=True)
    datum = db.Column(db.Date, nullable=False, unique=True)
    year_day_of_year = db.Column(db.String(255), nullable=False, unique=True)
    year_month = db.Column(db.String(255), nullable=False)
    year_week = db.Column(db.String(255), nullable=False)
    #
    year = db.Column(db.Integer, nullable=False)
    month = db.Column(db.Integer, nullable=False)
    day_of_month = db.Column(db.Integer, nullable=False)
    day_of_week = db.Column(db.Integer, nullable=False)
    day_of_year = db.Column(db.Integer, nullable=True)
    week_of_year = db.Column(db.Integer, nullable=False)
    #
    processed_update = db.Column(db.Boolean, nullable=False)
    processed_full_update = db.Column(db.Boolean, nullable=False)

    @classmethod
    def remove_all(cls):
        num_rows_deleted = 0
        try:
            num_rows_deleted = db.session.query(cls).delete()
            db.session.commit()
        except Exception:
            db.session.rollback()
        return num_rows_deleted


class RkiDatenstand(BlueprintDateReported):
    __tablename__ = 'rki_date_datenstand'
    __mapper_args__ = {'concrete': True}
    __table_args__ = (
        db.UniqueConstraint(
            'date_reported_import_str',
            'datum',
            'year_day_of_year',
            name="uix_rki_date_datenstand"
        ),
    )

    def __repr__(self):
        return "%s(%s, %s, %s)" % (self.__class__.__name__,
                                   self.date_reported_import_str, self.datum.isoformat(), self.year_day_of_year)

    id = db.Column(db.Integer, primary_key=True)
    #
    date_reported_import_str = db.Column(db.String(255), nullable=False, unique=True)
    datum = db.Column(db.Date, nullable=False, unique=True)
    year_day_of_year = db.Column(db.String(255), nullable=False, unique=True)
    year_month = db.Column(db.String(255), nullable=False)
    year_week = db.Column(db.String(255), nullable=False)
    #
    year = db.Column(db.Integer, nullable=False)
    month = db.Column(db.Integer, nullable=False)
    day_of_month = db.Column(db.Integer, nullable=False)
    day_of_week = db.Column(db.Integer, nullable=False)
    day_of_year = db.Column(db.Integer, nullable=True)
    week_of_year = db.Column(db.Integer, nullable=False)
    #
    processed_update = db.Column(db.Boolean, nullable=False)
    processed_full_update = db.Column(db.Boolean, nullable=False)


class RkiRefDatum(BlueprintDateReported):
    __tablename__ = 'rki_date_ref_datum'
    __mapper_args__ = {'concrete': True}
    __table_args__ = (
        db.UniqueConstraint(
            'date_reported_import_str',
            'datum',
            'year_day_of_year',
            name="uix_rki_date_ref_datum"
        ),
    )

    def __repr__(self):
        return "%s(%s, %s, %s)" % (self.__class__.__name__,
                                   self.date_reported_import_str, self.datum.isoformat(), self.year_day_of_year)

    id = db.Column(db.Integer, primary_key=True)
    #
    date_reported_import_str = db.Column(db.String(255), nullable=False, unique=True)
    datum = db.Column(db.Date, nullable=False, unique=True)
    year_day_of_year = db.Column(db.String(255), nullable=False, unique=True)
    year_month = db.Column(db.String(255), nullable=False)
    year_week = db.Column(db.String(255), nullable=False)
    #
    year = db.Column(db.Integer, nullable=False)
    month = db.Column(db.Integer, nullable=False)
    day_of_month = db.Column(db.Integer, nullable=False)
    day_of_week = db.Column(db.Integer, nullable=False)
    day_of_year = db.Column(db.Integer, nullable=True)
    week_of_year = db.Column(db.Integer, nullable=False)
    #
    processed_update = db.Column(db.Boolean, nullable=False)
    processed_full_update = db.Column(db.Boolean, nullable=False)


class RkiBundesland(BlueprintLocationGroup):
    __tablename__ = 'rki_landkreis_bundesland'
    __mapper_args__ = {'concrete': True}
    __table_args__ = (
        db.UniqueConstraint('id_bundesland', 'location_group', name="uix_rki_landkreis_bundesland"),
    )

    def __repr__(self):
        return "%s(%s %s)" % (self.__class__.__name__, self.location_group, self.id_bundesland)

    def __str__(self):
        return " " + self.location_group + " "

    id = db.Column(db.Integer, primary_key=True)
    id_bundesland = db.Column(db.String(255), nullable=False)
    location_group = db.Column(db.String(255), nullable=False)
    processed_update = db.Column(db.Boolean, nullable=False)
    processed_full_update = db.Column(db.Boolean, nullable=False)


class RkiLandkreis(BlueprintLocation):
    __tablename__ = 'rki_landkreis'
    __mapper_args__ = {'concrete': True}
    __table_args__ = (
        db.UniqueConstraint('location', 'id_landkreis', name="uix_rki_landkreis"),
    )

    def __repr__(self):
        return "%s(%s %s)" % (self.__class__.__name__, self.location, self.id_landkreis)

    def __str__(self):
        return " " + str(self.location_group) + " : " + self.location + " ( " + self.id_landkreis + " ) "

    id = db.Column(db.Integer, primary_key=True)
    processed_update = db.Column(db.Boolean, nullable=False)
    processed_full_update = db.Column(db.Boolean, nullable=False)
    location = db.Column(db.String(255), nullable=False)
    location_group_id = db.Column(db.Integer, db.ForeignKey('rki_landkreis_bundesland.id'), nullable=False)
    location_group = db.relationship(
        'RkiBundesland',
        lazy='joined',
        cascade='save-update',
        order_by='RkiBundesland.location_group')
    id_landkreis = db.Column(db.String(255), nullable=False)
    location_type = db.Column(db.String(255), nullable=False)

    @classmethod
    @cache.memoize(50)
    def get_bochum(cls):
        return db.session.query(cls)\
            .filter(cls.location == 'SK Bochum')\
            .one()

    @classmethod
    def find_all(cls):
        return db.session.query(cls).order_by(cls.location).all()

    @classmethod
    @cache.memoize(50)
    def find_by_location_group(cls, location_group: RkiBundesland):
        return db.session.query(cls).filter(
            cls.location_group == location_group
        ).order_by(cls.location).all()

    @classmethod
    @cache.memoize(50)
    def get_by_location_group(cls, location_group: RkiBundesland, page: int):
        return db.session.query(cls).filter(
            cls.location_group == location_group
        ).order_by(cls.location).paginate(page, per_page=ITEMS_PER_PAGE)

    @classmethod
    @cache.memoize(50)
    def find_all_as_dict(cls):
        rki_landkreis_dict = {}
        for my_rki_landkreis in cls.find_all():
            my_key = my_rki_landkreis.location_type + " " + my_rki_landkreis.location
            rki_landkreis_dict[my_key] = my_rki_landkreis
        return rki_landkreis_dict


class RkiAltersgruppe(BlueprintEntity):
    __tablename__ = 'rki_altersgruppe'
    __mapper_args__ = {'concrete': True}
    __table_args__ = (
        db.UniqueConstraint('altersgruppe', name="uix_rki_altersgruppe"),
    )

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, self.altersgruppe)

    def __str__(self):
        return " " + self.altersgruppe + " "

    id = db.Column(db.Integer, primary_key=True)
    processed_update = db.Column(db.Boolean, nullable=False)
    processed_full_update = db.Column(db.Boolean, nullable=False)
    altersgruppe = db.Column(db.String(255), nullable=False, unique=True)

    @classmethod
    @cache.memoize(50)
    def get_all_as_page(cls, page: int):
        return db.session.query(cls) \
            .order_by(cls.altersgruppe)\
            .paginate(page, per_page=ITEMS_PER_PAGE)

    @classmethod
    @cache.memoize(50)
    def get_all(cls, page: int):
        return db.session.query(cls).order_by(cls.altersgruppe).paginate(page, per_page=ITEMS_PER_PAGE)

    @classmethod
    @cache.memoize(50)
    def find_all(cls):
        return db.session.query(cls).order_by(cls.altersgruppe).all()

    @classmethod
    def find_all_as_dict(cls):
        altersgruppe_dict = {}
        for my_altersgruppe in cls.find_all():
            altersgruppe_dict[my_altersgruppe.altersgruppe] = my_altersgruppe
        return altersgruppe_dict


class RkiData(BlueprintFactTable):
    __tablename__ = 'rki'
    __mapper_args__ = {'concrete': True}
    __table_args__ = (
        db.UniqueConstraint(
            'fid',
            name="uix_rki"),
    )

    def __repr__(self):
        return "%s(%s %s %s %s %s %s %s)" % (self.__class__.__name__,
                                             self.fid, self.geschlecht, self.date_reported_id,
                                             self.datenstand_id, self.location_id,
                                             self.ref_datum_id, self.altersgruppe_id)

    id = db.Column(db.Integer, primary_key=True)
    processed_update = db.Column(db.Boolean, nullable=False)
    processed_full_update = db.Column(db.Boolean, nullable=False)
    #
    date_reported_id = db.Column(db.Integer, db.ForeignKey('rki_date_reported.id'), nullable=False)
    date_reported = db.relationship(
        'RkiMeldedatum',
        lazy='joined',
        cascade='save-update',
        order_by='desc(RkiMeldedatum.datum)')
    location_id = db.Column(db.Integer, db.ForeignKey('rki_landkreis.id'), nullable=False)
    location = db.relationship(
        'RkiLandkreis',
        lazy='joined',
        cascade='save-update',
        order_by='asc(RkiLandkreis.location)')
    #
    fid = db.Column(db.String(255), nullable=False)
    # altersgruppe = db.Column(db.String(255), nullable=False)
    altersgruppe_id = db.Column(db.Integer, db.ForeignKey('rki_altersgruppe.id'), nullable=False)
    altersgruppe = db.relationship(
        'RkiAltersgruppe',
        lazy='joined',
        cascade='save-update',
        order_by='desc(RkiAltersgruppe.altersgruppe)')
    # neuer_fall = db.Column(db.String(255), nullable=False)
    geschlecht = db.Column(db.String(255), nullable=False)
    anzahl_fall = db.Column(db.String(255), nullable=False)
    anzahl_todesfall = db.Column(db.String(255), nullable=False)
    # meldedatum = db.Column(db.String(255), nullable=False)
    # datenstand = db.Column(db.String(255), nullable=False)
    datenstand_id = db.Column(db.Integer, db.ForeignKey('rki_date_datenstand.id'), nullable=False)
    datenstand = db.relationship(
        'RkiDatenstand',
        lazy='joined',
        cascade='save-update',
        order_by='desc(RkiDatenstand.datum)')
    neuer_fall = db.Column(db.String(255), nullable=False)
    neuer_todesfall = db.Column(db.String(255), nullable=False)
    # ref_datum = db.Column(db.String(255), nullable=False)
    ref_datum_id = db.Column(db.Integer, db.ForeignKey('rki_date_ref_datum.id'), nullable=False)
    ref_datum = db.relationship(
        'RkiRefDatum',
        lazy='joined',
        cascade='save-update',
        order_by='desc(RkiRefDatum.datum)')
    neu_genesen = db.Column(db.String(255), nullable=False)
    anzahl_genesen = db.Column(db.String(255), nullable=False)
    ist_erkrankungsbeginn = db.Column(db.String(255), nullable=False)
    altersgruppe2 = db.Column(db.String(255), nullable=False)

    @classmethod
    def delete_all(cls):
        db.session.query(cls).delete()
        db.session.commit()
        return None

    @classmethod
    def __query_by_location(cls, location: RkiLandkreis):
        return db.session.query(cls).filter(
            cls.location_id == location.id
        ).populate_existing().options(
            joinedload(cls.location).joinedload(RkiLandkreis.location_group),
            joinedload(cls.date_reported)
        )

    @classmethod
    def __query_by_date_reported(cls, date_reported: RkiMeldedatum):
        return db.session.query(cls).filter(
            cls.date_reported_id == date_reported.id
        ).populate_existing().options(
            joinedload(cls.location).joinedload(RkiMeldedatum.location_group),
            joinedload(cls.date_reported)
        )

    @classmethod
    @cache.memoize(50)
    def get_by_location(cls, location: RkiLandkreis, page: int):
        return cls.__query_by_location(location).paginate(page, per_page=ITEMS_PER_PAGE)

    @classmethod
    @cache.memoize(50)
    def find_by_location(cls, location: RkiLandkreis):
        return cls.__query_by_location(location).all()

    @classmethod
    @cache.memoize(50)
    def find_by_date_reported_and_location(cls, date_reported: RkiMeldedatum, location: RkiLandkreis):
        return db.session.query(cls)\
            .filter(and_((cls.date_reported_id == date_reported.id), (cls.location_id == location.id)))\
            .all()

    @classmethod
    @cache.memoize(50)
    def get_by_date_reported_and_location(cls, date_reported: RkiMeldedatum, location: RkiLandkreis, page: int):
        return db.session.query(cls)\
            .filter(and_((cls.date_reported_id == date_reported.id), (cls.location_id == location.id)))\
            .paginate(page, per_page=ITEMS_PER_PAGE)

    @classmethod
    def delete_data_for_one_day(cls, date_reported: RkiMeldedatum):
        for data in cls.find_by_date_reported(date_reported):
            db.session.delete(data)
        db.session.delete(date_reported)
        db.session.commit()
