from sqlalchemy.orm import Bundle
from database import db, cache
from flask_covid19.blueprints.app_all.all_model import AllImport, AllFlat


class WhoImport(AllImport):
    __tablename__ = 'who_import'
    __mapper_args__ = {'concrete': True}

    def __repr__(self):
        return "%s(%s %s %s %s %s)" % (self.__class__.__name__,
                                    self.datum.isoformat(), self.date_reported,
                                    self.country_code, self.country, self.who_region)

    def __str__(self):
        return self.datum.isoformat() + " " + self.country_code + " " + self.country + " " + str(self.row_imported)

    id = db.Column(db.Integer, primary_key=True)
    datum = db.Column(db.Date, nullable=False)
    processed_update = db.Column(db.Boolean, nullable=False)
    processed_full_update = db.Column(db.Boolean, nullable=False)
    #
    new_cases = db.Column(db.String(255), nullable=False)
    cumulative_cases = db.Column(db.String(255), nullable=False)
    new_deaths = db.Column(db.String(255), nullable=False)
    cumulative_deaths = db.Column(db.String(255), nullable=False)
    country_code = db.Column(db.String(255), nullable=False)
    country = db.Column(db.String(255), nullable=False)
    who_region = db.Column(db.String(255), nullable=False)
    date_reported = db.Column(db.String(255), nullable=False)

    @classmethod
    @cache.cached(timeout=50)
    def get_regions(cls):
        return db.session.query(cls.who_region)\
            .order_by(cls.who_region)\
            .distinct().all()

    @classmethod
    @cache.cached(timeout=50)
    def get_dates_reported(cls):
        return db.session.query(cls.date_reported)\
            .order_by(cls.date_reported.desc())\
            .distinct().all()

    @classmethod
    @cache.memoize(50)
    def get_for_one_day(cls, day: str):
        return db.session.query(cls)\
            .filter(cls.date_reported == day)\
            .order_by(cls.country.asc())\
            .all()

    @classmethod
    def get_dates_reported_as_string_array(cls):
        myresultarray = []
        myresultset = db.session.query(cls.datum)\
            .order_by(cls.datum.desc()).group_by(cls.datum)\
            .distinct().all()
        for my_datum_item in myresultset:
            my_datum = my_datum_item.datum.isoformat()
            if my_datum not in myresultarray:
                myresultarray.append(my_datum)
        return myresultarray

    @classmethod
    def __get_new_dates_as_array_sql(cls):
        # TODO: #83 WhoImport.get_new_dates_as_array() SQLalchemy instead of SQL
        sql_query = """
            select
                distinct 
                    who_import.date_reported
                from
                    who_import
                where
                    date_reported
                not in (
                    select
                        distinct
                            who_date_reported.date_reported_import_str
                        from
                            who
                        left join
                            who_date_reported
                        on
                            who.date_reported_id=who_date_reported.id
                        group by 
                            who_date_reported.date_reported_import_str
                        order by
                            who_date_reported.date_reported_import_str desc
                )
                group by
                    who_import.date_reported
                order by 
                    who_import.date_reported desc
            """
        new_dates = []
        for item in db.session.execute(sql_query):
            new_dates.append(item['date_reported'])
        return new_dates

    @classmethod
    def __get_new_dates_as_array_orm(cls):
        # TODO: #83 WhoImport.get_new_dates_as_array() SQLalchemy instead of SQL
        sql_query = """
                select
                    distinct 
                        who_import.date_reported
                    from
                        who_import
                    where
                        date_reported
                    not in (
                        select
                            distinct
                                who_date_reported.date_reported_import_str
                            from
                                who
                            left join
                                who_date_reported
                            on
                                who.date_reported_id=who_date_reported.id
                            group by 
                                who_date_reported.date_reported_import_str
                            order by
                                who_date_reported.date_reported_import_str desc
                    )
                    group by
                        who_import.date_reported
                    order by 
                        who_import.date_reported desc
                """
        new_dates = []
        for item in db.session.execute(sql_query):
            new_dates.append(item['date_reported'])
        return new_dates

    @classmethod
    @cache.cached(timeout=50)
    def get_new_dates_as_array(cls):
        # return cls.__get_new_dates_as_array_sql()
        return cls.__get_new_dates_as_array_orm()

    @classmethod
    @cache.cached(timeout=50)
    def countries(cls):
        bu = Bundle('countries', cls.country_code, cls.country, cls.who_region)
        return db.session.query(bu).distinct()

    @classmethod
    @cache.cached(timeout=50)
    def get_datum_of_all_who_import(cls):
        dates_reported = []
        bu = Bundle('dates_reported', cls.date_reported)
        for date_reported in db.session.query(bu).distinct().order_by(cls.date_reported):
            item = date_reported[0][0]
            if not item in dates_reported:
                dates_reported.append(item)
        return dates_reported


class WhoFlat(AllFlat):
    __tablename__ = 'who_import_flat'
    __mapper_args__ = {'concrete': True}

    def __repr__(self):
        return "%s(%s %s %s %s %s)" % (self.__class__.__name__,
                                       self.datum.isoformat(), self.date_reported_import_str,
                                       self.location_code, self.location, self.location_group)

    def __str__(self):
        return self.datum.isoformat() + " " + self.location_code + " " + self.location + " " + str(self.location_group)

    id = db.Column(db.Integer, primary_key=True)
    #
    date_reported_import_str = db.Column(db.String(255), nullable=False)
    datum = db.Column(db.Date, nullable=False)
    #
    year = db.Column(db.Integer, nullable=False)
    year_month = db.Column(db.String(255), nullable=False)
    year_week = db.Column(db.String(255), nullable=False)
    year_day_of_year = db.Column(db.String(255), nullable=False)
    processed_update = db.Column(db.Boolean, nullable=False)
    processed_full_update = db.Column(db.Boolean, nullable=False)
    #
    month = db.Column(db.Integer, nullable=False)
    day_of_month = db.Column(db.Integer, nullable=False)
    day_of_week = db.Column(db.Integer, nullable=False)
    week_of_year = db.Column(db.Integer, nullable=False)
    day_of_year = db.Column(db.Integer, nullable=False)
    location = db.Column(db.String(255), nullable=False)
    location_group = db.Column(db.String(255), nullable=False)
    location_code = db.Column(db.String(255), nullable=False)
    #
    new_cases = db.Column(db.Integer, nullable=False)
    cumulative_cases = db.Column(db.Integer, nullable=False)
    new_deaths = db.Column(db.Integer, nullable=False)
    cumulative_deaths = db.Column(db.Integer, nullable=False)
    country_code = db.Column(db.String(255), nullable=False)
    country = db.Column(db.String(255), nullable=False)
    who_region = db.Column(db.String(255), nullable=False)
    date_reported = db.Column(db.String(255), nullable=False)
